/* hexbuf(char *dest,void *bin,int bytes):
 * convert some binary data to hex, usually for printing.
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. [Omitted.]
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include <stdio.h>
#include "afflib.h"
#include <err.h>
#include <string.h>
#include <stdlib.h>
#include <zlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <openssl/rand.h>
#include <arpa/inet.h>

/* Utility Functions */

struct quad {
    unsigned long low;
    unsigned long high;
};



const char *af_hexbuf(char *dst,const unsigned char *bin,int bytes)
{
    const char *start = dst;		// remember where the start of the string is

    while(bytes>0){
	sprintf(dst,"%02x",*bin); // convert the next byte
	dst += 2;
	bin += 1;
	bytes--;
    }
    return start;			// return the start
}




/****************************************************************
 *** low-level functions for putting...
 ****************************************************************/

void af_put_segv(AFFILE *af, const char *name,unsigned long arg,const void *data,int datalen)
{
    struct af_segment_head segh;
    struct af_segment_tail segt;

    assert(sizeof(segh)==16);
    assert(sizeof(segt)==8);

    int namelen = strlen(name);

    strcpy(segh.flag,AF_SEGHEAD);
    segh.name_len = htonl(namelen);
    segh.data_len = htonl(datalen);
    segh.arg      = htonl(arg);

    strcpy(segt.flag,AF_SEGTAIL);
    segt.segment_len = htonl(sizeof(segh)+namelen + datalen + sizeof(segt));
    
    if(fwrite(&segh,sizeof(segh),1,af->aseg)!=1) err(1,"af_put_seg - head");
    if(fwrite(name,1,namelen,af->aseg)!=namelen) err(1,"af_put_seg - name");
    if(fwrite(data,1,datalen,af->aseg)!=datalen) err(1,"af_put_seg - data");
    if(fwrite(&segt,sizeof(segt),1,af->aseg)!=1) err(1,"af_put_seg - tail");
}

void af_put_segq(AFFILE *af,const char *name,unsigned long long value)
{
    struct quad  q;
    q.low  = htonl(value & 0xffffffff);
    q.high = htonl(value >> 32);
    
    af_put_segv(af,name,0,&q,8);
}

/****************************************************************
 *** Probe Functions 
 ****************************************************************/

int af_probe_next_segv(AFFILE *af,char **segname,unsigned long *datasize,unsigned long *segsize)
{
    if(!af->aseg) err(1,"af_probe_segment only works with aff files");    

    struct af_segment_head segh;
    off_t start = ftello(af->aseg);

    if(fread(&segh,sizeof(segh),1,af->aseg)!=1){
	return -1;
    }
    unsigned long name_len = ntohl(segh.name_len);
    unsigned long data_len = ntohl(segh.data_len);
    *segname = (char *)malloc(name_len+1);
    if(fread(*segname,1,name_len,af->aseg)!=name_len){
	fseeko(af->aseg,start,SEEK_SET); // rewind
	return -1;
    }
    (*segname)[name_len] = 0;
    fseeko(af->aseg,start,SEEK_SET); // rewind

    if(datasize){
	*datasize = data_len;
    }
    if(segsize){
	*segsize = sizeof(struct af_segment_head) + sizeof(struct af_segment_tail)
	    + name_len + data_len;
    }
    return 0;
}




/****************************************************************
 *** GET FUNCTIONS
 ****************************************************************/

/****************************************************************
 *** Probe the next segment:
 *** Return its name and argument, but don't advance the pointer....
 *** Returns 0 on success, -1 on end of file or other error.
 ****************************************************************/
/****************************************************************
 *** Get the next segment.
 *** segname must be freed;
 *** data must be freed.
 *** if no data is specified, then just seek past it...
 *** Returns 0 on success, -1 on end of file or other error.
 ****************************************************************/
int af_get_next_segv(AFFILE *af,char **segname,unsigned long *arg,
			unsigned char **data,size_t *datalen)
{
    if(!af->aseg) err(1,"af_get_next_segv only works with aff files");

    struct af_segment_head segh;
    struct af_segment_tail segt;
    if(fread(&segh,sizeof(segh),1,af->aseg)!=1){
	/* end of file trying to read segment. */
	return -1;
    }
    if(strcmp(segh.flag,AF_SEGHEAD)!=0){
	fprintf(stderr,"af_get_next_segv: AF file is corrupt (bad head)!\n");
	exit(1);
    }
    int name_len = ntohl(segh.name_len);
    int data_len = ntohl(segh.data_len);
    if(arg) *arg = ntohl(segh.arg);

    *segname = (char *)malloc(name_len+1);
    if(fread(*segname,1,name_len,af->aseg)!=name_len){
	/* Oops. Shoudln't get an EOF here... */
	fprintf(stderr,"af_get_next_segv: end of file reading segment name...\n");
	return -1;
    }
    (*segname)[name_len] = 0;		// null-terminate the name
    if(data){
	/* Allocate the data and read... */
	*data    = (unsigned char *)malloc(data_len);
	if(fread(*data,1,data_len,af->aseg)!=data_len){
	    return -1;			// end of file!
	}
    } else {
	fseeko(af->aseg,data_len,SEEK_CUR); // skip past the data
    }
    if(datalen) *datalen = data_len;
    if(fread(&segt,1,sizeof(segt),af->aseg)!=sizeof(segt)){
	fprintf(stderr,"af_get_next_segv: end of file reading segment tail...\n");
	return -1;
    }
    unsigned long stl = ntohl(segt.segment_len);
    if(strcmp(segt.flag,AF_SEGTAIL)!=0 ||
       stl != sizeof(segh)+name_len + data_len + sizeof(segt)){
	fprintf(stderr,"af_get_next_segv: AF file is corrupt reading tail!\n");
	//exit(1);
    }
    return 0;
}


/* af_get_segment:
 * Get the named segment.
 * If it is the next segment, just get it.
 * Otherwise go to the beginning and scan for it.
 * Eventually we hope to cache the names and starting locations of each segment...
 */

int af_get_segv(AFFILE *af,const char *name,unsigned long *arg,
		unsigned char **data,size_t *datalen)
{
    char *next;
    int first = 1;
    unsigned long segsize = 0;
    
    do {
	if(af_probe_next_segv(af,&next,0,&segsize)==0){	
	    if(strcmp(next,name)==0){	// found the segment!
		free(next);
		int ret = af_get_next_segv(af,&next,arg,data,datalen);
		assert(strcmp(next,name)==0);	// hopefully it's still the same!
		free(next);			// still don't need it anymore
		return ret;
	    }
	}
	else {
	    /* Must be at EOF. If this is not first time through,
	     * then segment does not exist...
	     */
	    if(!first) break;
	}
	if(first){
	    /* If first time we checked, rewind to the first segment */
	    af_rewind(af);
	    first = 0;			// no longer the first time through
	    continue;
	}
	fseeko(af->aseg,segsize,SEEK_CUR); // seek to next segment
    } while(1);
    return -1;				// couldn't find segment
}





/* get_segment:
 * Get a segment. If it is compressed, uncompress it.
 * *data will point to the uncompressed segment.
 * if shouldfree is set, then data should be freed.
 * Return 0 if success, -1 if fail.
 */

int af_get_image_segment(AFFILE *af,int segnum,
			 unsigned char **data,size_t *bytes)
{
    char segname[256];
    unsigned long arg=0;
    unsigned char *segdata;
    unsigned long segdata_len;

    sprintf(segname,AF_SEG,segnum);
    if(af_get_segv(af,segname,&arg,&segdata,&segdata_len)){
	return -1;			// couldn't get it
    }
    
    /* See if we need to uncompress it */
    if((arg & AF_SEG_COMPRESSED)==0){
	/* It's not compressed; just return the raw segment. */
	*data = segdata;
	*bytes = segdata_len;
	return 0;
    }

    /* Need to uncompress it */
    *data = (unsigned char *)malloc(af->segsize);
    *bytes = af->segsize;

    uLongf uncompressed_bytes=af->segsize;

    int r = uncompress(*data,&uncompressed_bytes,(const Bytef *)segdata,segdata_len);
    switch(r){
    case Z_OK:
	*bytes = uncompressed_bytes;
	free(segdata);		// don't need this one anymore
	return 0;
    case Z_MEM_ERROR:
	err(1,"Z_MEM_ERROR decompressing segment %d",segnum);
    case Z_DATA_ERROR:
	err(2,"Z_DATA_ERROR decompressing segment %d",segnum);
    default:
	err(3,"uncompress returned an invalid value in get_segment");
    }
    return -1;				// should never get here!
}

int af_get_segq(AFFILE *af,const char *name,unsigned long long *quad)
{
    unsigned char *buf;
    unsigned long  bufsize;
    struct quad *q;
    if(af_get_segv(af,name,0,&buf,&bufsize)){
	return -1;			// couldn't get it...
    }
    if(bufsize!=sizeof(*q)){		// make sure size is good.
	free(buf);
	return -1;
    }
    q = (struct quad *)buf;		// point q to buf.
	
    *quad =
	(((unsigned long long)ntohl(q->high)) << 32) +
	((unsigned long long)ntohl(q->low));
    free(buf);				// free allocated stuff.
    return 0;
}

void af_get_segsize(AFFILE *af)
{
    if(af_get_segv(af,AF_SEGSIZE,&af->segsize,0,0)) err(1,"Can't get segsize");
    if(af_get_segq(af,AF_IMAGESIZE,&af->imagesize)) err(1,"Can't get imagesize");
}




/****************************************************************
 *** Update functions
 ****************************************************************/
int af_rewind(AFFILE *af)
{
    fseeko(af->aseg,sizeof(struct af_head),SEEK_SET); // go to the beginning
    return 0;
}

int af_backspace(AFFILE *af)
{
    struct af_segment_head segh;
    struct af_segment_tail segt;


    off_t start = ftello(af->aseg);
    off_t pos_tail = start - sizeof(segt); // backspace to read the tail
    fseeko(af->aseg,pos_tail,SEEK_SET);
    if(fread(&segt,1,sizeof(segt),af->aseg)!=1){
	fseeko(af->aseg,start,SEEK_SET); // put it back
	return -1;			// can't read segt?
    }
    /* Now I know how long the segment was. Compute where it started */
    off_t seg_start = start - ntohl(segt.segment_len);
    fseeko(af->aseg,seg_start,SEEK_SET);
    return 0;
}

int af_update_segv(AFFILE *af, const char *name,
		    unsigned long arg,const void *value,int vallen)
{
    char *next;
    unsigned long segsize = 0;
    unsigned long datasize = 0;

    af_rewind(af);			// start at the beginning
    while(af_probe_next_segv(af,&next,&datasize,&segsize)==0){
	if(strcmp(next,name)==0){	// found the segment!
	    free(next);
	    if(datasize != vallen){	// wrong size; can't update
		err(1,"af_update_segv: %d != %d when updating segment '%s'",
		    datasize,vallen,name);
	    }
	    /* We can just write in place! */
	    af_put_segv(af,name,arg,value,vallen);
	    return 0;
	}
	else {
	    fseeko(af->aseg,segsize,SEEK_CUR); // skip this segment
	}
    }
    return -1;				// couldn't find segment
}


int af_update_segq(AFFILE *af, const char *name, unsigned long long value)
{
    struct quad  q;
    q.low  = htonl(value & 0xffffffff);
    q.high = htonl(value >> 32);
    
    return af_update_segv(af,name,0,&q,8);
}


/****************************************************************
 *** AFFILE routines
 ****************************************************************/

#ifndef O_BINARY
#define O_BINARY 0			// for Windows compatability
#endif

AFFILE *af_open(const char *filename,int flags,int mode)
{
    if(flags & O_WRONLY) err(1,"af_open: flag O_WRONLY not supported");

    int fd = open(filename,flags | O_BINARY,mode);
    if(fd<0) return 0;			// couldn't open

    AFFILE *af = (AFFILE *)calloc(sizeof(AFFILE),1);

    /* Figure out the extension; if it is not a '.aff' extension, treat
     * as a raw file...*/
    char *cc = strrchr(filename,'.');
    if(cc==0 || strcmp(cc,".aff")!=0){
	af->raw=fdopen(fd,"rb");

	struct stat sb;
	if(stat(filename,&sb)){
	    err(1,"stat(%s): ",filename);
	}
	af->imagesize = sb.st_size;
	return af;
    }
	
    /* If we are creating, we must be creating an AFF... */
    char strflag[8];
    strcpy(strflag,"r");		// we have to be able to read
    if(flags & O_RDWR) 	strcpy(strflag,"w+");

    af->aseg = fdopen(fd,strflag);
    if(!af->aseg) err(1,"fdopen(%d,%s)",fd,strflag);

    /* Oh, we are reading. See if it is an AF header */
    char buf[8];
    if(fread(buf,1,sizeof(buf),af->aseg)!=sizeof(buf)){
	/* Hm. End of file. Okay. Rewind to the beginning and put in a proper header */
	fseeko(af->aseg,0,SEEK_SET);
	fwrite(AF_HEADER,1,8,af->aseg);	// writes the NUL too!a
	return af;
    }

    if(strcmp(buf,AF_HEADER)!=0){
	buf[7] = 0;
	fprintf(stderr,"%s is not an AFF file (header=%s)\n",filename,buf);
	exit(1);
    }
    
    af->segnum    = -1;			// no segment loaded
    return af;
}

/* Return if we are at the end of the file */
int af_eof(AFFILE *af)
{
    if(af->raw) return feof(af->raw);
    if(af->segsize==0) af_get_segsize(af);
    return af->pos >= af->imagesize;
}

/* Close the image */
int af_close(AFFILE *af)
{
    if(af->raw)  fclose(af->raw);
    if(af->aseg) fclose(af->aseg);
    memset(af,0,sizeof(*af));
    free(af);
    return 0;
}

/* Throw out the current segment */
int af_purge(AFFILE *af)
{
    if(af->aseg){
	free(af->segbuf);
	af->segbuf = 0;
	af->segnum = -1;
    }
    return 0;
}


#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

/* Read from the affile */
int af_read(AFFILE *af,unsigned char *buf,int count)
{
    if(af->raw){	/* passthrough? */
	return fread(buf,1,count,af->raw);
    }

    /* We need to have a segsize to continue */
    if(af->segsize==0) af_get_segsize(af);

    int total = 0;
    while(count>0){
	/* If no segment is loaded, load the current segment */
	if(af->segnum == -1){
	    af->segnum = af->pos / af->segsize;	// will be the segment we want
	    if(af_get_image_segment(af,af->segnum,&af->segbuf, &af->segbytes)){
		break;			// no more to get
	    }
	}
	int seg_offset = af->pos - af->segnum * af->segsize; // where we were reading from
	int seg_left   = af->segbytes - seg_offset; // number we can get out
	int bytes_to_read = MIN(count,seg_left); // how many we can copy out

	assert(bytes_to_read >= 0);	// 
	if(bytes_to_read==0) return total; // that's all we could get

	/* Copy out the bytes for the user */
	memcpy(buf,af->segbuf+seg_offset,bytes_to_read); // copy out
	buf     += bytes_to_read;
	af->pos += bytes_to_read;
	count   -= bytes_to_read;
	total   += bytes_to_read;

	/* If we read out all of the bytes that were left in the segment,
	 * then we are at the end of the segment, so throw it away.
	 */
	if(seg_left == bytes_to_read){	
	    af_purge(af);
	}
    }
    /* We have copied all of the user's requested data, so return */
    return total;
}

/* Seek in the virtual file */
fpos_t af_seek(AFFILE *af,fpos_t pos,int whence)
{
    if(af->raw) return fseeko(af->raw,pos,whence);

    switch(whence){
    case SEEK_SET:
	af->pos = pos;
	break;
    case SEEK_CUR:
	af->pos += pos;
	break;
    case SEEK_END:
	af->pos = af->imagesize - pos;
	break;
    }
    if(af->pos < 0) af->pos = 0;
    if(af->pos > af->imagesize) af->pos = af->imagesize;

    /* If the correct segment is not loaded, purge the segment */
    int new_seg = af->pos / af->segsize;
    if(new_seg != af->segnum){
	af_purge(af);
    }

    return af->pos;
}

fpos_t  af_tell(AFFILE *af)
{
    if(af->raw) return ftell(af->raw);
    return af->pos;
}


/****************************************************************
 *** AFF creation functions
 ****************************************************************/

void af_put_badflag(AFFILE *af)
{
    RAND_pseudo_bytes(af->badflag,sizeof(af->badflag));
    strcpy((char *)af->badflag,"BAD SECTOR");

    af_put_segv(af,AF_BADFLAG,0,af->badflag,sizeof(af->badflag));
    af_put_segv(af,AF_BADSECTORS,0,0,0);
}


void af_setsize(AFFILE *af,long segsize,unsigned long long imagesize)
{
    if(af->segsize){
	fprintf(stderr,"af_setsize: segsize already set to %d!\n",af->segsize);
	exit(1);
    }
    af_put_segv(af,AF_SEGSIZE,segsize,0,0);
    af_put_segq(af,AF_IMAGESIZE,imagesize);
}

